==> tiledb-cloud-0.10.7.tar.gz: tiledb-cloud-0.10.7/src/tiledb/cloud/bioimg/ingestion.py <==
import math
import os
from typing import Any, Dict, Iterator, Mapping, Optional, Sequence, Tuple, Union

import tiledb
from tiledb.cloud._common.utils import logger

DEFAULT_RESOURCES = {"cpu": "8", "memory": "4Gi"}
DEFAULT_IMG_NAME = "3.9-imaging-dev"


def ingest(
    source: Union[Sequence[str], str],
    output: str,
    config: Mapping[str, Any],
    *args: Any,
    taskgraph_name: Optional[str] = None,
    num_batches: Optional[int] = None,
    threads: Optional[int] = 8,
    resources: Optional[Mapping[str, Any]] = None,
    namespace: Optional[str],
    **kwargs,
) -> tiledb.cloud.dag.DAG:
    """The function ingests microscopy images into TileDB arrays

    :param source: uri / iterable of uris of input files
    :param output: output dir for the ingested tiledb arrays
    :param config: dict configuration to pass on tiledb.VFS
    :param taskgraph_name: Optional name for taskgraph, defaults to None
    :param num_batches: Number of graph nodes to spawn.
        Performs it sequentially if default, defaults to 1
    :param threads: Number of threads for node side multiprocessing, defaults to 8
    :param resources: configuration for node specs e.g. {"cpu": "8", "memory": "4Gi"},
        defaults to None
    :param namespace: The namespace where the DAG will run
    """

    def ingest_tiff_udf(
        io_uris: Sequence[Tuple],
        config: Mapping[str, Any],
        workers: int,
        *args: Any,
        **kwargs,
    ):
        """Internal udf that ingests server side batch of bioimaging files
        into tiledb arrays using tiledb-bioimg API

        :param io_uris: Pairs of tiff input - output tdb uris
        :param config: dict configuration to pass on tiledb.VFS
        :param workers: Number of threads that will spawn for parallelizing ingestion
        """

        from tiledb.bioimg.converters.ome_tiff import OMETiffConverter

        conf = tiledb.Config(params=config)
        vfs = tiledb.VFS(config=conf)

        with tiledb.scope_ctx(ctx_or_config=conf):
            for input, output in io_uris:
                with vfs.open(input) as src:
                    OMETiffConverter.to_tiledb(
                        src, output, *args, max_workers=workers, chunked=True, **kwargs
                    )

    if isinstance(source, str):
        # Handle only lists
        source = [source]

    # Calculate number batches - default fair split

    def batch(iterable, chunks):
        length = len(iterable)
        for ndx in range(0, length, chunks):
            yield iterable[ndx : min(ndx + chunks, length)]

    # If num_batches is default create number of images nodes
    # constraint node max_workers to 20 fully heuristic
    if num_batches is None:
        num_batches = len(source)
        batch_size = 1
        max_workers = 20
    else:
        batch_size = math.ceil(len(source) / num_batches)
        max_workers = None

    # Get the list of all BioImg samples input/out
    samples = get_uris(source, output, config)

    # Build the task graph
    dag_name = "batch-ingest-bioimg" if taskgraph_name is None else taskgraph_name
    task_prefix = f"{dag_name}  - Batch Task"

    logger.info("Building graph")
    graph = tiledb.cloud.dag.DAG(
        name=dag_name,
        mode=tiledb.cloud.dag.Mode.BATCH,
        max_workers=max_workers,
        namespace=namespace,
    )

    for i, work in enumerate(batch(samples, batch_size)):
        logger.info(f"Adding batch {i}")
        graph.submit(
            ingest_tiff_udf,
            work,
            config,
            threads,
            *args,
            name=f"{task_prefix} - {i}/{num_batches}",
            mode=tiledb.cloud.dag.Mode.BATCH,
            resources=DEFAULT_RESOURCES if resources is None else resources,
            image_name=DEFAULT_IMG_NAME,
            **kwargs,
        )

    graph.compute()
    return graph


def ingest_udf(*args: Any, **kwargs: Any) -> Dict[str, str]:
    """Ingestor wrapper function that can be used as a UDF."""
    grf = ingest(*args, **kwargs)
    return {"status": "started", "graph_id": str(grf.server_graph_uuid)}


def get_uris(source: Sequence[str], output_dir: str, config: Mapping[str, Any]):
    """Match input uri/s with output destinations

    :param source: A sequence of paths or path to input
    :param output_dir: A path to the output directory
    """
    vfs = tiledb.VFS(config=config)

    def create_output_path(input_file, output_dir) -> str:
        return os.path.join(output_dir, os.path.basename(input_file) + ".tdb")

    def iter_paths(sequence) -> Iterator[Tuple]:
        for uri in sequence:
            yield uri, create_output_path(uri, output_dir)

    if len(source) == 1 and vfs.is_dir(source[0]):
        # Folder like input
        return tuple(iter_paths(vfs.ls(source[0])))
    elif isinstance(source, Sequence):
        # List of input uris - single file is one element list
        return tuple(iter_paths(source))

==> tiledb-cloud-0.10.9.tar.gz: tiledb-cloud-0.10.9/src/tiledb/cloud/bioimg/ingestion.py <==
import math
import os
from typing import Any, Dict, Iterator, Mapping, Optional, Sequence, Tuple, Union

import tiledb
from tiledb.cloud._common.utils import logger

DEFAULT_RESOURCES = {"cpu": "8", "memory": "4Gi"}
DEFAULT_IMG_NAME = "3.9-imaging-dev"


def ingest(
    source: Union[Sequence[str], str],
    output: str,
    config: Mapping[str, Any],
    *args: Any,
    taskgraph_name: Optional[str] = None,
    num_batches: Optional[int] = None,
    threads: Optional[int] = 8,
    resources: Optional[Mapping[str, Any]] = None,
    namespace: Optional[str],
    **kwargs,
) -> tiledb.cloud.dag.DAG:
    """The function ingests microscopy images into TileDB arrays

    :param source: uri / iterable of uris of input files
    :param output: output dir for the ingested tiledb arrays
    :param config: dict configuration to pass on tiledb.VFS
    :param taskgraph_name: Optional name for taskgraph, defaults to None
    :param num_batches: Number of graph nodes to spawn.
        Performs it sequentially if default, defaults to 1
    :param threads: Number of threads for node side multiprocessing, defaults to 8
    :param resources: configuration for node specs e.g. {"cpu": "8", "memory": "4Gi"},
        defaults to None
    :param namespace: The namespace where the DAG will run
    """

    def ingest_tiff_udf(
        io_uris: Sequence[Tuple],
        config: Mapping[str, Any],
        *args: Any,
        **kwargs,
    ):
        """Internal udf that ingests server side batch of bioimaging files
        into tiledb arrays using tiledb-bioimg API

        :param io_uris: Pairs of tiff input - output tdb uris
        :param config: dict configuration to pass on tiledb.VFS
        """

        from tiledb import filter
        from tiledb.bioimg.converters.ome_tiff import OMETiffConverter

        compressor = kwargs.get("compressor", None)
        if compressor:
            compressor_args = dict(compressor)
            compressor_name = compressor_args.pop("_name")
            if compressor_name:
                compressor_args = {
                    k: None if not v else v
                    for k, v in compressor.get("attrs", {}).items()
                }
                kwargs["compressor"] = vars(filter).get(compressor_name)(
                    **compressor_args
                )
            else:
                raise ValueError

        conf = tiledb.Config(params=config)
        vfs = tiledb.VFS(config=conf)

        with tiledb.scope_ctx(ctx_or_config=conf):
            for input, output in io_uris:
                with vfs.open(input) as src:
                    OMETiffConverter.to_tiledb(src, output, **kwargs)

    if isinstance(source, str):
        # Handle only lists
        source = [source]

    # Calculate number batches - default fair split

    def batch(iterable, chunks):
        length = len(iterable)
        for ndx in range(0, length, chunks):
            yield iterable[ndx : min(ndx + chunks, length)]

    # If num_batches is default create number of images nodes
    # constraint node max_workers to 20 fully heuristic
    if num_batches is None:
        num_batches = len(source)
        batch_size = 1
        max_workers = 20
    else:
        batch_size = math.ceil(len(source) / num_batches)
        max_workers = None

    # Get the list of all BioImg samples input/out
    samples = get_uris(source, output, config)

    # Build the task graph
    dag_name = "bioimg-ingestion" if taskgraph_name is None else taskgraph_name
    task_prefix = f"{dag_name} - Task"

    logger.info("Building graph")
    graph = tiledb.cloud.dag.DAG(
        name=dag_name,
        mode=tiledb.cloud.dag.Mode.BATCH,
        max_workers=max_workers,
        namespace=namespace,
    )

    # serialize udf arguments
    compressor = kwargs.pop("compressor", None)
    compressor_serial = serialize_filter(compressor) if compressor else None

    for i, work in enumerate(batch(samples, batch_size)):
        logger.info(f"Adding batch {i}")
        graph.submit(
            ingest_tiff_udf,
            work,
            config,
            threads,
            *args,
            name=f"{task_prefix} - {i}/{num_batches}",
            mode=tiledb.cloud.dag.Mode.BATCH,
            resources=DEFAULT_RESOURCES if resources is None else resources,
            image_name=DEFAULT_IMG_NAME,
            compressor=compressor_serial,
            **kwargs,
        )

    graph.compute()
    return graph


def serialize_filter(filter):
    if isinstance(filter, tiledb.Filter):
        filter_dict = filter._attrs_()
        filter_dict["_name"] = type(filter).__name__
        return filter_dict
    else:
        raise TypeError


def ingest_udf(*args: Any, **kwargs: Any) -> Dict[str, str]:
    """Ingestor wrapper function that can be used as a UDF."""
    grf = ingest(*args, **kwargs)
    return {"status": "started", "graph_id": str(grf.server_graph_uuid)}


def get_uris(source: Sequence[str], output_dir: str, config: Mapping[str, Any]):
    """Match input uri/s with output destinations

    :param source: A sequence of paths or path to input
    :param output_dir: A path to the output directory
    """
    vfs = tiledb.VFS(config=config)

    def create_output_path(input_file, output_dir) -> str:
        return os.path.join(output_dir, os.path.basename(input_file) + ".tdb")

    def iter_paths(sequence) -> Iterator[Tuple]:
        for uri in sequence:
            yield uri, create_output_path(uri, output_dir)

    if len(source) == 1 and vfs.is_dir(source[0]):
        # Folder like input
        return tuple(iter_paths(vfs.ls(source[0])))
    elif isinstance(source, Sequence):
        # List of input uris - single file is one element list
        return tuple(iter_paths(source))
